/**
 * Copyright (c) 2010-2014, Zoltan Ujhelyi, IncQuery Labs Ltd.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Zoltan Ujhelyi - initial API and implementation
 */
package org.eclipse.viatra.query.patternlanguage.typing;

import com.google.common.base.Objects;
import com.google.inject.Inject;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.viatra.query.patternlanguage.helper.CorePatternLanguageHelper;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.AggregatedValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.AggregatorExpression;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.BoolValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.CheckConstraint;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.CompareConstraint;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.CompareFeature;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.CountAggregator;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.DoubleValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Expression;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.FunctionEvaluationValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.IntValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.ListValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.PathExpressionConstraint;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.PathExpressionHead;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.PathExpressionTail;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Pattern;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.PatternCall;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.PatternCompositionConstraint;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.StringValue;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Type;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.ValueReference;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Variable;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.VariableReference;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.VariableValue;
import org.eclipse.viatra.query.patternlanguage.typing.ITypeSystem;
import org.eclipse.viatra.query.patternlanguage.typing.TypeInformation;
import org.eclipse.viatra.query.patternlanguage.typing.judgements.ParameterTypeJudgement;
import org.eclipse.viatra.query.patternlanguage.typing.judgements.TypeConformJudgement;
import org.eclipse.viatra.query.patternlanguage.typing.judgements.TypeJudgement;
import org.eclipse.viatra.query.patternlanguage.typing.judgements.XbaseExpressionTypeJudgement;
import org.eclipse.viatra.query.runtime.matchers.context.IInputKey;
import org.eclipse.viatra.query.runtime.matchers.context.common.JavaTransitiveInstancesKey;
import org.eclipse.xtext.xbase.XExpression;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.eclipse.xtext.xbase.typesystem.IBatchTypeResolver;

/**
 * @author Zoltan Ujhelyi
 * @since 1.3
 */
@SuppressWarnings("all")
public class PatternLanguageTypeRules {
  @Inject
  private ITypeSystem typeSystem;
  
  @Inject
  private IBatchTypeResolver typeResolver;
  
  protected void _inferTypes(final Pattern pattern, final TypeInformation information) {
    EList<Variable> _parameters = pattern.getParameters();
    final Procedure1<Variable> _function = new Procedure1<Variable>() {
      @Override
      public void apply(final Variable parameter) {
        Type _type = parameter.getType();
        boolean _notEquals = (!Objects.equal(_type, null));
        if (_notEquals) {
          Type _type_1 = parameter.getType();
          final IInputKey typeKey = PatternLanguageTypeRules.this.typeSystem.extractTypeDescriptor(_type_1);
          information.declareType(parameter, typeKey);
        }
        Set<Variable> _localReferencesOfParameter = CorePatternLanguageHelper.getLocalReferencesOfParameter(parameter);
        final Procedure1<Variable> _function = new Procedure1<Variable>() {
          @Override
          public void apply(final Variable ref) {
            TypeConformJudgement _typeConformJudgement = new TypeConformJudgement(ref, parameter);
            information.provideType(_typeConformJudgement);
            TypeConformJudgement _typeConformJudgement_1 = new TypeConformJudgement(parameter, ref);
            information.provideType(_typeConformJudgement_1);
          }
        };
        IterableExtensions.<Variable>forEach(_localReferencesOfParameter, _function);
      }
    };
    IterableExtensions.<Variable>forEach(_parameters, _function);
  }
  
  protected void _inferTypes(final CheckConstraint constraint, final TypeInformation information) {
  }
  
  protected void _inferTypes(final CompareConstraint constraint, final TypeInformation information) {
    CompareFeature _feature = constraint.getFeature();
    boolean _equals = Objects.equal(_feature, CompareFeature.EQUALITY);
    if (_equals) {
      ValueReference _leftOperand = constraint.getLeftOperand();
      ValueReference _rightOperand = constraint.getRightOperand();
      TypeConformJudgement _typeConformJudgement = new TypeConformJudgement(_leftOperand, _rightOperand);
      information.provideType(_typeConformJudgement);
      ValueReference _rightOperand_1 = constraint.getRightOperand();
      ValueReference _leftOperand_1 = constraint.getLeftOperand();
      TypeConformJudgement _typeConformJudgement_1 = new TypeConformJudgement(_rightOperand_1, _leftOperand_1);
      information.provideType(_typeConformJudgement_1);
    }
  }
  
  protected void _inferTypes(final PatternCompositionConstraint constraint, final TypeInformation information) {
    final PatternCall call = constraint.getCall();
    final Pattern pattern = call.getPatternRef();
    for (int i = 0; (i < Math.min(call.getParameters().size(), pattern.getParameters().size())); i++) {
      EList<ValueReference> _parameters = call.getParameters();
      ValueReference _get = _parameters.get(i);
      EList<Variable> _parameters_1 = pattern.getParameters();
      Variable _get_1 = _parameters_1.get(i);
      ParameterTypeJudgement _parameterTypeJudgement = new ParameterTypeJudgement(_get, _get_1);
      information.provideType(_parameterTypeJudgement);
    }
  }
  
  protected void _inferTypes(final PathExpressionConstraint constraint, final TypeInformation information) {
    PathExpressionHead _head = constraint.getHead();
    Type _type = _head.getType();
    final IInputKey sourceType = this.typeSystem.extractTypeDescriptor(_type);
    PathExpressionHead _head_1 = constraint.getHead();
    PathExpressionTail tail = _head_1.getTail();
    while ((!Objects.equal(tail.getTail(), null))) {
      PathExpressionTail _tail = tail.getTail();
      tail = _tail;
    }
    Type _type_1 = tail.getType();
    final IInputKey targetType = this.typeSystem.extractTypeDescriptor(_type_1);
    if (((!Objects.equal(sourceType, null)) && (!Objects.equal(targetType, null)))) {
      PathExpressionHead _head_2 = constraint.getHead();
      VariableReference _src = _head_2.getSrc();
      TypeJudgement _typeJudgement = new TypeJudgement(_src, sourceType);
      information.provideType(_typeJudgement);
      PathExpressionHead _head_3 = constraint.getHead();
      ValueReference _dst = _head_3.getDst();
      TypeJudgement _typeJudgement_1 = new TypeJudgement(_dst, targetType);
      information.provideType(_typeJudgement_1);
    }
  }
  
  protected void _inferTypes(final AggregatedValue reference, final TypeInformation information) {
    AggregatorExpression _aggregator = reference.getAggregator();
    boolean _matched = false;
    if (_aggregator instanceof CountAggregator) {
      _matched=true;
      JavaTransitiveInstancesKey _javaTransitiveInstancesKey = new JavaTransitiveInstancesKey(Integer.class);
      TypeJudgement _typeJudgement = new TypeJudgement(reference, _javaTransitiveInstancesKey);
      information.provideType(_typeJudgement);
    }
    if (!_matched) {
      AggregatorExpression _aggregator_1 = reference.getAggregator();
      EClass _eClass = _aggregator_1.eClass();
      String _name = _eClass.getName();
      String _plus = ("Unsupported aggregator type " + _name);
      throw new IllegalArgumentException(_plus);
    }
  }
  
  protected void _inferTypes(final Expression reference, final TypeInformation information) {
  }
  
  protected void _inferTypes(final FunctionEvaluationValue reference, final TypeInformation information) {
    XExpression _expression = reference.getExpression();
    XbaseExpressionTypeJudgement _xbaseExpressionTypeJudgement = new XbaseExpressionTypeJudgement(reference, _expression, this.typeResolver);
    information.provideType(_xbaseExpressionTypeJudgement);
  }
  
  protected void _inferTypes(final BoolValue reference, final TypeInformation information) {
    JavaTransitiveInstancesKey _javaTransitiveInstancesKey = new JavaTransitiveInstancesKey(Boolean.class);
    TypeJudgement _typeJudgement = new TypeJudgement(reference, _javaTransitiveInstancesKey);
    information.provideType(_typeJudgement);
  }
  
  protected void _inferTypes(final DoubleValue reference, final TypeInformation information) {
    JavaTransitiveInstancesKey _javaTransitiveInstancesKey = new JavaTransitiveInstancesKey(Double.class);
    TypeJudgement _typeJudgement = new TypeJudgement(reference, _javaTransitiveInstancesKey);
    information.provideType(_typeJudgement);
  }
  
  protected void _inferTypes(final IntValue reference, final TypeInformation information) {
    JavaTransitiveInstancesKey _javaTransitiveInstancesKey = new JavaTransitiveInstancesKey(Integer.class);
    TypeJudgement _typeJudgement = new TypeJudgement(reference, _javaTransitiveInstancesKey);
    information.provideType(_typeJudgement);
  }
  
  protected void _inferTypes(final ListValue reference, final TypeInformation information) {
    JavaTransitiveInstancesKey _javaTransitiveInstancesKey = new JavaTransitiveInstancesKey(List.class);
    TypeJudgement _typeJudgement = new TypeJudgement(reference, _javaTransitiveInstancesKey);
    information.provideType(_typeJudgement);
  }
  
  protected void _inferTypes(final StringValue reference, final TypeInformation information) {
    JavaTransitiveInstancesKey _javaTransitiveInstancesKey = new JavaTransitiveInstancesKey(String.class);
    TypeJudgement _typeJudgement = new TypeJudgement(reference, _javaTransitiveInstancesKey);
    information.provideType(_typeJudgement);
  }
  
  protected void _inferTypes(final VariableValue reference, final TypeInformation information) {
  }
  
  public void inferTypes(final EObject reference, final TypeInformation information) {
    if (reference instanceof AggregatedValue) {
      _inferTypes((AggregatedValue)reference, information);
      return;
    } else if (reference instanceof BoolValue) {
      _inferTypes((BoolValue)reference, information);
      return;
    } else if (reference instanceof DoubleValue) {
      _inferTypes((DoubleValue)reference, information);
      return;
    } else if (reference instanceof FunctionEvaluationValue) {
      _inferTypes((FunctionEvaluationValue)reference, information);
      return;
    } else if (reference instanceof IntValue) {
      _inferTypes((IntValue)reference, information);
      return;
    } else if (reference instanceof ListValue) {
      _inferTypes((ListValue)reference, information);
      return;
    } else if (reference instanceof StringValue) {
      _inferTypes((StringValue)reference, information);
      return;
    } else if (reference instanceof VariableValue) {
      _inferTypes((VariableValue)reference, information);
      return;
    } else if (reference instanceof CheckConstraint) {
      _inferTypes((CheckConstraint)reference, information);
      return;
    } else if (reference instanceof CompareConstraint) {
      _inferTypes((CompareConstraint)reference, information);
      return;
    } else if (reference instanceof PathExpressionConstraint) {
      _inferTypes((PathExpressionConstraint)reference, information);
      return;
    } else if (reference instanceof PatternCompositionConstraint) {
      _inferTypes((PatternCompositionConstraint)reference, information);
      return;
    } else if (reference instanceof Expression) {
      _inferTypes((Expression)reference, information);
      return;
    } else if (reference instanceof Pattern) {
      _inferTypes((Pattern)reference, information);
      return;
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(reference, information).toString());
    }
  }
}
